// Old Turbo-C CONIO.H compatibility library for LCC-Win32
// and GCC/EGCS Mingw32 compilers.
// Version 1.0 (September 1999).
// Created by Daniel Guerrero Miralles (daguer@geocities.com).
// This source code is public domain.

#include <Python.h>
#include <Windows.h>
#include <malloc.h>
#include <conio.h>
#include "WConio.h"

#define NORM_ATTR LIGHTGRAY
#define NORM_CURSOR_SIZE 10
#define SOLID_CURSOR_SIZE 99

struct module_state {
    PyObject *WConioError;
};

#if PY_MAJOR_VERSION >= 3
#define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
#else
#define GETSTATE(m) (&_state)
static struct module_state _state;
#endif

// INTERNAL FUNCTIONS /////////////////////////////////////////////////////


static HANDLE WINAPI GetConHandle (TCHAR * pszName)
{
	SECURITY_ATTRIBUTES sa;

	sa.nLength = sizeof (sa);
	sa.lpSecurityDescriptor = NULL;
	sa.bInheritHandle = TRUE;

	// Using CreateFile we get the true console handle, avoiding
	// any redirection.

	return CreateFile (pszName,
		GENERIC_READ|GENERIC_WRITE,
		FILE_SHARE_READ|FILE_SHARE_WRITE,
		&sa,
		OPEN_EXISTING,
		(DWORD) 0,
		(HANDLE) 0);
}

#define GetConOut() GetConHandle(TEXT("CONOUT$"))
#define GetConIn() GetConHandle(TEXT("CONIN$"))
#define ReleaseConOut(h) CloseHandle(h)
#define ReleaseConIn(h) CloseHandle(h)

// clear the screen
static int int_clrscr (void)
{
	HANDLE hConOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	COORD coord;
	DWORD dwDummy;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		if (GetConsoleScreenBufferInfo (hConOut, &csbi)) {
			coord.X = 0;
			coord.Y = 0;
			FillConsoleOutputCharacter (hConOut, (TCHAR)' ', 
                csbi.dwSize.X * csbi.dwSize.Y, coord, &dwDummy);
			FillConsoleOutputAttribute (hConOut, csbi.wAttributes, 
                csbi.dwSize.X * csbi.dwSize.Y, coord, &dwDummy);
			SetConsoleCursorPosition (hConOut, coord);
		}
		ReleaseConOut (hConOut);

        return 0;
	}

    return -1;
}


// set the text foreground/background colors
static int int_textattr(int newattr)
{
	HANDLE hConOut;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		SetConsoleTextAttribute (hConOut, (WORD) newattr);
		ReleaseConOut (hConOut);
        return 0;
	}

    return -1;
}


// set the cursor shape
static void int_setcursortype(int cur_t)
{
	HANDLE hConOut;
	CONSOLE_CURSOR_INFO cci;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		switch (cur_t)
		{
			case _NOCURSOR:
				cci.dwSize = NORM_CURSOR_SIZE;
				cci.bVisible = FALSE;
				break;
			case _SOLIDCURSOR:
				cci.dwSize = SOLID_CURSOR_SIZE;
				cci.bVisible = TRUE;
				break;
			case _NORMALCURSOR:
				cci.dwSize = NORM_CURSOR_SIZE;
				cci.bVisible = TRUE;
				break;
		}
		SetConsoleCursorInfo (hConOut, &cci);
		ReleaseConOut (hConOut);
	}
}


static BOOL WINAPI GetScreenInfo(PCONSOLE_SCREEN_BUFFER_INFO pCSBI)
{
	HANDLE hOut;
	BOOL fSuccess;

	hOut = GetConOut();
	if (hOut != INVALID_HANDLE_VALUE) {
		fSuccess = GetConsoleScreenBufferInfo (hOut, pCSBI);
		ReleaseConOut (hOut);
		return fSuccess;
	}
	return FALSE;
}


static void WINAPI ConPuts(const char * pszText) // unused
{
	HANDLE hConOut;
	DWORD dwDummy;
	DWORD dwLen;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE) {
		dwLen = lstrlenA (pszText);
		WriteConsoleA (hConOut, pszText, dwLen, &dwDummy, NULL);
		ReleaseConOut (hConOut);
	}
}


// CONTROL FUNCTIONS //////////////////////////////////////////////////////

static PyObject *
gettextinfo(PyObject *self, PyObject *args)
{
	CONSOLE_SCREEN_BUFFER_INFO csbi;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;
    
	if(GetScreenInfo (&csbi))
        return Py_BuildValue("(i,i,i,i,i,i,i,i,i,i,i)",
		    csbi.srWindow.Left,
		    csbi.srWindow.Top,
		    csbi.srWindow.Right,
		    csbi.srWindow.Bottom,
		    csbi.wAttributes,
		    NORM_ATTR, C80,
		    csbi.dwSize.Y,
		    csbi.dwSize.X,
		    csbi.dwCursorPosition.X,
		    csbi.dwCursorPosition.Y);

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetScreenInfo Failed"));
    return NULL;
}


// CURSOR FUNCTIONS ///////////////////////////////////////////////////////

// get cursor horizontal possition
static PyObject *
wherex(PyObject *self, PyObject *args)
{
	CONSOLE_SCREEN_BUFFER_INFO csbi;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	if(GetScreenInfo(&csbi))
		return Py_BuildValue("i", csbi.dwCursorPosition.X);

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetScreenInfo Failed"));
    return NULL;
}


// get cursor vertical possition
static PyObject *
wherey(PyObject *self, PyObject *args)
{
	CONSOLE_SCREEN_BUFFER_INFO csbi;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	if (GetScreenInfo(&csbi))
		return Py_BuildValue("i", csbi.dwCursorPosition.Y);

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetScreenInfo Failed"));
    return NULL;
}

// Change the cursor location
static PyObject *
gotoxy(PyObject *self, PyObject *args)
{
	HANDLE hOut;
	COORD coord;
    int x, y;

    if (!PyArg_ParseTuple(args, "ii", &x, &y))
        return NULL;

	hOut = GetConOut();
	if (hOut != INVALID_HANDLE_VALUE)
	{
		coord.X = x;
		coord.Y = y;
		SetConsoleCursorPosition(hOut, coord);
		ReleaseConOut(hOut);
        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


static PyObject *
setcursortype(PyObject *self, PyObject *args)
{
    int cur_t;

    if (!PyArg_ParseTuple(args, "i", &cur_t))
        return NULL;

    switch(cur_t) {
    case 0:
        int_setcursortype(_NOCURSOR);
        break;
    case 1:
        int_setcursortype(_NORMALCURSOR);
        break;
    case 2:
        int_setcursortype(_SOLIDCURSOR);
        break;
    }

    Py_INCREF(Py_None);
    return Py_None;
}


// change text color to high intensity
static PyObject *
highvideo(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		if (GetConsoleScreenBufferInfo(hConOut, &csbi)) {
			SetConsoleTextAttribute (hConOut, 
                csbi.wAttributes | FOREGROUND_INTENSITY);
		}
		ReleaseConOut(hConOut);

        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}

// change text color to low intensity
static PyObject *
lowvideo(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		if (GetConsoleScreenBufferInfo (hConOut, &csbi)) {
			SetConsoleTextAttribute(hConOut, 
                csbi.wAttributes & (~FOREGROUND_INTENSITY));
		}
		ReleaseConOut(hConOut);

        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


static PyObject *
textattr(PyObject *self, PyObject *args)
{
    int attr;

    if (!PyArg_ParseTuple(args, "i", &attr))
        return NULL;

    if(int_textattr(attr) != 0) {
        PyErr_SetObject(GETSTATE(self)->WConioError, 
            Py_BuildValue("s", "textattr Failed"));
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}


// TEXT I/O FUNCTIONS /////////////////////////////////////////////////////

//clear up to end-of-line
static PyObject *
clreol(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	DWORD dwDummy;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		if (GetConsoleScreenBufferInfo(hConOut, &csbi)) {
			FillConsoleOutputCharacter(hConOut, (TCHAR)' ', 
                csbi.dwSize.X - csbi.dwCursorPosition.X, 
                csbi.dwCursorPosition, &dwDummy);
			FillConsoleOutputAttribute(hConOut, csbi.wAttributes, 
                csbi.dwSize.X - csbi.dwCursorPosition.X, 
                csbi.dwCursorPosition, &dwDummy);
		}
		ReleaseConOut(hConOut);

        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


// clear the screen
static PyObject *
clrscr(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	if(int_clrscr() != 0) {
        PyErr_SetObject(GETSTATE(self)->WConioError, 
            Py_BuildValue("s", "clrscr Failed"));
        return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
}


// delete current line
static PyObject *
delline(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	SMALL_RECT srSource;
	COORD dwDest;
	CHAR_INFO ciFill;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		if (GetConsoleScreenBufferInfo (hConOut, &csbi))
		{
			srSource.Top = csbi.dwCursorPosition.Y + 1;
			srSource.Left = csbi.srWindow.Left;
			srSource.Bottom = csbi.srWindow.Bottom;
			srSource.Right = csbi.srWindow.Right;
			dwDest.X = csbi.srWindow.Left;
			dwDest.Y = csbi.dwCursorPosition.Y;
			// FIXME!
			// this may not work in an big-endian machine.
			ciFill.Char.UnicodeChar = (WCHAR)' ';
			ciFill.Attributes = csbi.wAttributes;
			if (!ScrollConsoleScreenBuffer (hConOut, &srSource, NULL, dwDest, &ciFill)) {
				PyErr_SetFromWindowsErr(0);
				ReleaseConOut (hConOut);
				return NULL;
			};
		}
		ReleaseConOut (hConOut);

        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


// insert a line
static PyObject *
insline(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	SMALL_RECT srSource;
	COORD dwDest;
	CHAR_INFO ciFill;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE)
	{
		if (GetConsoleScreenBufferInfo (hConOut, &csbi))
		{
			srSource.Top = csbi.dwCursorPosition.Y;
			srSource.Left = csbi.srWindow.Left;
			srSource.Bottom = csbi.srWindow.Bottom;
			srSource.Right = csbi.srWindow.Right;
			dwDest.X = csbi.srWindow.Left;
			dwDest.Y = csbi.dwCursorPosition.Y + 1;
			// FIXME!
			// This may not work in a big-endian machine.
			ciFill.Char.UnicodeChar = (WCHAR)' ';
			ciFill.Attributes = csbi.wAttributes;
			ScrollConsoleScreenBuffer (hConOut, &srSource, NULL, dwDest, &ciFill);
		}
		ReleaseConOut (hConOut);

        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


// get text from a rectangular section of the screen
static PyObject *
gettext(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	PCHAR_INFO pCI;
	COORD dwBufferSize, dwBufferOrg;
	SMALL_RECT srSource;
	DWORD dwBuffLen, i;
	char *destin = NULL;
    char *pchDest;
    int left, top, right, bottom;
    PyObject *result;

    if (!PyArg_ParseTuple(args, "iiii", &left, &top, &right, &bottom))
        return NULL;

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE) {
		srSource.Left = left;
		srSource.Top = top;
		srSource.Right = right;
		srSource.Bottom = bottom;
		dwBufferSize.X = srSource.Right - srSource.Left + 1;
		dwBufferSize.Y = srSource.Bottom - srSource.Top + 1;
		dwBufferOrg.X = 0;
		dwBufferOrg.Y = 0;
		dwBuffLen = dwBufferSize.X * dwBufferSize.Y;

		pCI = malloc (sizeof (CHAR_INFO) * dwBuffLen);
		if (pCI != NULL) {
			if (ReadConsoleOutputA (hConOut, pCI, dwBufferSize, 
                dwBufferOrg, &srSource))
			{
			    destin = malloc(dwBuffLen*2);
				pchDest = (char *) destin;
				// convert to the expected output format
				for (i = 0; i < dwBuffLen; i++) {
					*pchDest = (char) pCI[i].Char.AsciiChar;
					pchDest++;
					*pchDest = (char) pCI[i].Attributes;
					pchDest++;
				}
			}
			free (pCI);
		}
		ReleaseConOut (hConOut);
#if PY_MAJOR_VERSION >= 3
        result = Py_BuildValue("y#", destin, dwBuffLen*2);
#else  
        result = Py_BuildValue("s#", destin, dwBuffLen*2);
#endif 
        if (destin!=NULL) free(destin);
        return result;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


// Set text in a rectangular section of the screen
static PyObject *
puttext(PyObject *self, PyObject *args)
{
	HANDLE hConOut;
	PCHAR_INFO pCI;
	COORD dwBufferSize, dwBufferOrg;
	SMALL_RECT srDest;
	DWORD dwBuffLen, i;
	char *pchSource;
    int left, top, right, bottom, len;
    char *source;

#if PY_MAJOR_VERSION >= 3
    if (!PyArg_ParseTuple(args, "iiiiy#", 
        &left, &top, &right, &bottom, &source, &len))
            return NULL;
#else
    if (!PyArg_ParseTuple(args, "iiiis#", 
        &left, &top, &right, &bottom, &source, &len))
            return NULL;
#endif

	hConOut = GetConOut();
	if (hConOut != INVALID_HANDLE_VALUE) {
		srDest.Left = left;
		srDest.Top = top;
		srDest.Right = right; // coords are inclusive
		srDest.Bottom = bottom;
		dwBufferSize.X = srDest.Right - srDest.Left + 1;
		dwBufferSize.Y = srDest.Bottom - srDest.Top + 1;
		dwBufferOrg.X = 0;
		dwBufferOrg.Y = 0;
		dwBuffLen = dwBufferSize.X * dwBufferSize.Y;

        if((dwBuffLen * 2) > (DWORD)len) {
            PyErr_SetObject(GETSTATE(self)->WConioError, 
                Py_BuildValue("s", "String Too Short"));
            return NULL;
        }

		pCI = malloc (sizeof (CHAR_INFO) * dwBuffLen);
		if (pCI != NULL) {
			pchSource = source;
			for (i = 0; i < dwBuffLen; i++) {
				pCI[i].Char.AsciiChar = (CHAR) *pchSource;
				pchSource++;
				pCI[i].Attributes = (WORD) *pchSource;
				pchSource++;
			}
			WriteConsoleOutputA (hConOut, pCI, dwBufferSize, 
                dwBufferOrg, &srDest);
			free (pCI);
		}
		ReleaseConOut (hConOut);

        Py_INCREF(Py_None);
        return Py_None;
	}

    PyErr_SetObject(GETSTATE(self)->WConioError, 
        Py_BuildValue("s", "GetConOut Failed"));
    return NULL;
}


static PyObject *
settitle(PyObject *self, PyObject *args)
{
    PyObject *title;
    wchar_t *bufferu;
    char *bufferb;
    Py_ssize_t size, n;
    
    if (!PyArg_ParseTuple(args, "O:settitle", &title)) 
        goto error;

    if (PyUnicode_Check(title)) {
        size = PyUnicode_GetSize(title);
        bufferu = malloc((size+1) * sizeof(wchar_t));
        n = PyUnicode_AsWideChar((PyUnicodeObject*) title, bufferu, size);
        if (n<0) {
            free(bufferu);
            goto error;
        }
        bufferu[n] = (wchar_t) 0;
        if (!SetConsoleTitleW(bufferu)) {
            free(bufferu);
            goto error;
        }
        free(bufferu);
        goto ok;
    }
    else if (PyBytes_Check(title)) {
        bufferb = PyBytes_AS_STRING(title);
        if (!SetConsoleTitleA(bufferb)) 
            goto error;
        goto ok;
    }
    else if (PyByteArray_Check(title)) {
        bufferb = PyByteArray_AS_STRING(title);
        if (!SetConsoleTitleA(bufferb)) 
            goto error;
        goto ok;
    }

error:        
    PyErr_SetObject(GETSTATE(self)->WConioError,
        Py_BuildValue("s", "SetConsoleTitle Failed"));
    return NULL;

ok:
    Py_INCREF(Py_None);
    return Py_None;
}


static PyObject *
meth_kbhit(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    return Py_BuildValue("i", kbhit());
}


static PyObject *
meth_getch(PyObject *self, PyObject *args)
{
    wchar_t c;

    if (!PyArg_ParseTuple(args, ""))
        return NULL;

    c = _getwch();

    return Py_BuildValue("(iu#)", (int)c, &c, 1);
}


static PyObject *
meth_putch(PyObject *self, PyObject *args)
{
    PyObject *o;
    wchar_t c;

    if (!PyArg_ParseTuple(args, "O:putch", &o)) 
        goto error;

    if PyLong_Check(o) {
        c = (wchar_t) PyLong_AsLong(o);
#if PY_MAJOR_VERSION < 3
    } else if PyInt_Check(o) {
        c = (wchar_t) PyInt_AsLong(o);
        if (PyErr_Occurred()) 
            goto error;
#endif
    } else if PyUnicode_Check(o) {
        if (PyUnicode_GetSize(o)!=1) 
            goto error;
        PyUnicode_AsWideChar((PyUnicodeObject*) o, &c, 1);
    } else if PyBytes_Check(o) {
        if (PyBytes_Size(o)!=1) 
            goto error;
        c = (wchar_t) (*PyBytes_AS_STRING(o));
    } else if PyByteArray_Check(o) {
        if (PyByteArray_Size(o)!=1) 
            goto error;
        c = (wchar_t) (*PyByteArray_AS_STRING(o));
    } else 
        goto error;

    _putwch(c);

// ok:
    Py_INCREF(Py_None);
    return Py_None;
    
error:
    PyErr_SetObject(GETSTATE(self)->WConioError,
        Py_BuildValue("s", "putch Failed"));
    return NULL;
}


static PyObject *
meth_ungetch(PyObject *self, PyObject *args)
{
    PyObject *o;
    wchar_t c;

    if (!PyArg_ParseTuple(args, "O:putch", &o)) 
        goto error;

    if PyLong_Check(o) {
        c = (wchar_t) PyLong_AsLong(o);
    } else if PyUnicode_Check(o) {
        if (PyUnicode_GetSize(o)!=1) 
            goto error;
        PyUnicode_AsWideChar((PyUnicodeObject*) o, &c, 1);
    } else if PyBytes_Check(o) {
        if (PyBytes_Size(o)!=1) 
            goto error;
        c = (wchar_t) (*PyBytes_AS_STRING(o));
    } else if PyByteArray_Check(o) {
        if (PyByteArray_Size(o)!=1) 
            goto error;
        c = (wchar_t) (*PyByteArray_AS_STRING(o));
    }    

    _ungetwch(c);

// ok:
    Py_INCREF(Py_None);
    return Py_None;
    
error:
    PyErr_SetObject(GETSTATE(self)->WConioError,
        Py_BuildValue("s", "ungetch Failed"));
    return NULL;
}


static PyMethodDef WConioMethods[] = {
    {"gettextinfo",    (PyCFunction) gettextinfo,    METH_VARARGS},
    {"wherex",         (PyCFunction) wherex,         METH_VARARGS},
    {"wherey",         (PyCFunction) wherey,         METH_VARARGS},
    {"gotoxy",         (PyCFunction) gotoxy,         METH_VARARGS},
    {"highvideo",      (PyCFunction) highvideo,      METH_VARARGS},
    {"lowvideo",       (PyCFunction) lowvideo,       METH_VARARGS},
    {"textattr",       (PyCFunction) textattr,       METH_VARARGS},
    {"setcursortype",  (PyCFunction) setcursortype,  METH_VARARGS},
    {"clrscr",         (PyCFunction) clrscr,         METH_VARARGS},
    {"clreol",         (PyCFunction) clreol,         METH_VARARGS},
    {"insline",        (PyCFunction) insline,        METH_VARARGS},
    {"delline",        (PyCFunction) delline,        METH_VARARGS},
    {"gettext",        (PyCFunction) gettext,        METH_VARARGS},
    {"puttext",        (PyCFunction) puttext,        METH_VARARGS},
    {"settitle",       (PyCFunction) settitle,       METH_VARARGS},
    {"kbhit",          (PyCFunction) meth_kbhit,     METH_VARARGS},
    {"getch",          (PyCFunction) meth_getch,     METH_VARARGS},
    {"putch",          (PyCFunction) meth_putch,     METH_VARARGS},
    {"ungetch",        (PyCFunction) meth_ungetch,   METH_VARARGS},
    {NULL,             NULL,           0} /* Sentinel */
};

/* common initialization function */
static BOOL
really_init(PyObject *module)
{
  PyObject *dict;
  struct module_state *st = GETSTATE(module);
  
  st->WConioError = PyErr_NewException("_WConio.error", NULL, NULL);
  if (st->WConioError == NULL) 
        return FALSE;
  dict = PyModule_GetDict(module);
  PyDict_SetItemString(dict, "error", st->WConioError);
  return TRUE;
}

#if PY_MAJOR_VERSION >= 3

static int wconio_traverse(PyObject *m, visitproc visit, void *arg) {
    Py_VISIT(GETSTATE(m)->WConioError);
    return 0;
}

static int wconio_clear(PyObject *m) {
    Py_CLEAR(GETSTATE(m)->WConioError);
    return 0;
}

static struct PyModuleDef moduledef = {
        PyModuleDef_HEAD_INIT,
        "_WConio",
        NULL,
        sizeof(struct module_state),
        WConioMethods,
        NULL,
        wconio_traverse,
        wconio_clear,
        NULL
};

PyObject *
PyInit__WConio(void)
{
    PyObject *module = PyModule_Create(&moduledef);
    if (module == NULL)
        return NULL;
    if (!really_init(module)) {
        Py_DECREF(module);
        return NULL;
    }
    return module;
}

#else

void
init_WConio()
{
    PyObject *module;
    module = Py_InitModule("_WConio", WConioMethods);
    really_init(module);
}

#endif
// end of file.
